@@ -1,6 +1,7 @@
 from ._flink_tasks import FlinkTasks, in_parallel
+from ._context import TaskContext
 from ._serialisation import DefaultSerialiser
 from ._types import RetryPolicy, TaskAlreadyExistsException, Task, Group
 from ._pipeline import PipelineBuilder
 from .messages_pb2 import TaskRequest, TaskResult, TaskException, TaskActionRequest, TaskActionResult, TaskActionException, \
-    TaskAction, TaskStatus
+    TaskAction, TaskStatus, TaskDeferral, DeferredTask
