@@ -2,7 +2,7 @@ from ._serialisation import DefaultSerialiser
 from ._utils import _gen_id, _task_type_for, _is_tuple, _type_name, _annotated_protos_for
 from ._types import RetryPolicy, TaskAlreadyExistsException
 from ._pipeline import _Pipeline, PipelineBuilder
-from ._context import _TaskContext
+from ._context import TaskContext
 from ._builtins import run_pipeline
 from ._protobuf import _pack_any
 from ._task_actions import _invoke_task_action
@@ -12,7 +12,6 @@ from .messages_pb2 import TaskRequest, TaskResult, TaskException, TaskActionRequ
 from statefun.request_reply import BatchContext
 from datetime import timedelta
 from typing import Union
-from functools import partial
 import logging
 import traceback as tb
 import inspect
@@ -45,6 +44,7 @@ def _create_task_exception(task_input, ex):
             exception_message=str(ex),
             stacktrace=tb.format_exc())
 
+        # if the task failed then ensure that exception retains the state from the task input (i.e. the TaskRequest)
         if isinstance(task_input, TaskRequest) and task_input.HasField('state'):
             task_exception.state.CopyFrom(task_input.state)
 
@@ -62,9 +62,6 @@ def _create_task_result(task_input, result=None):
             id=task_input.id,
             type=f'{_task_name(task_input)}.result')
 
-        if task_input.HasField('state'):
-            task_result.state.CopyFrom(task_input.state)
-
     if result is not None:
         task_result.result.CopyFrom(_pack_any(result))
 
@@ -92,7 +89,7 @@ class FlinkTasks(object):
 
         self._bindings = {}
 
-        self.register_builtin('run_pipeline', partial(run_pipeline, serialiser=self._serialiser))
+        self.register_builtin('run_pipeline', run_pipeline)
 
     def register_builtin(self, type_name, fun, **params):
         """
@@ -105,12 +102,13 @@ class FlinkTasks(object):
         """
         self._bindings[f'__builtins.{type_name}'] = _FlinkTask(fun, self._serialiser, **params)
 
-    def register(self, fun, module_name=None, **params):
+    def register(self, fun, wrapper=None, module_name=None, **params):
         """
         Registers a Python function as a Flink Task.
         Normally you would attribute the function with @tasks.bind() instead
 
         :param fun: the python function
+        :param optional wrapper: if wrapping a task function with e.g. functools.wraps then pass the wrapper here
         :param optional module_name: the module name to register the task under which by default is the Python module name containing the function
         :param params: any additional parameters to the Flink Task (such as a retry policy)
         """
@@ -119,20 +117,20 @@ class FlinkTasks(object):
             raise ValueError("function instance must be provided")
 
         fun.type_name = _task_type_for(fun, module_name)
-        self._bindings[fun.type_name] = _FlinkTask(fun, self._serialiser, **params)
+        self._bindings[fun.type_name] = _FlinkTask(wrapper or fun, self._serialiser, **params)
 
-    def bind(self, namespace: str = None, worker_name: str = None, retry_policy: RetryPolicy = None,  with_state: bool = False, is_fruitful: bool = True, module_name: str = None):
+    def bind(self, namespace: str = None, worker_name: str = None, retry_policy: RetryPolicy = None,  with_state: bool = False, is_fruitful: bool = True, module_name: str = None, with_context: bool = False):
         """
         Binds a function as a Flink Task
 
         :param namespace: namespace to use in place of the default
         :param worker_name: worker name to use in place of the default
         :param retry_policy: retry policy to use should the task throw an exception
-        :param with_state: whether to pass a state object as the first parameter - return value should be a tuple of state, result (default False)
+        :param with_state: whether to pass a state object as the first (second if with_context is also set) parameter.  The return value should be a tuple of state, result (default False)
         :param is_fruitful: whether the function produces a fruitful result or simply returns None (default True)
-        :param module_name: if specified then the task type used in addressingwill be module_name.function_name
+        :param module_name: if specified then the task type used in addressing will be module_name.function_name
                             otherwise the Python module containing the function will be used
-
+        :param with_context: whether to pass a Flink context object as the first parameter (default false)
         """
         def wrapper(function):
             def defaults():
@@ -142,7 +140,8 @@ class FlinkTasks(object):
                     'retry_policy': None if retry_policy is None else retry_policy.to_proto(),
                     'with_state': with_state,
                     'is_fruitful': is_fruitful,
-                    'module_name': module_name
+                    'module_name': module_name,
+                    'with_context': with_context
                 }
 
             def send(*args, **kwargs):
@@ -189,7 +188,7 @@ class FlinkTasks(object):
         :param task_input: the task input protobuf message
         """
 
-        with _TaskContext(context, _task_name(task_input), self._egress_type_name, self._serialiser) as task_context:
+        with TaskContext(context, _task_name(task_input), self._egress_type_name, self._serialiser) as task_context:
             try:
                 _log.info(f'Started {task_context}')
 
@@ -215,7 +214,7 @@ class FlinkTasks(object):
         :param context: context object provided by Flink
         :param task_input: the task input protobuf message
         """
-        with _TaskContext(context, _task_name(task_input), self._egress_type_name, self._serialiser) as task_context:
+        with TaskContext(context, _task_name(task_input), self._egress_type_name, self._serialiser) as task_context:
             try:
                 _log.info(f'Started {task_context}')
 
@@ -268,6 +267,7 @@ class FlinkTasks(object):
         elif isinstance(task_input, TaskRequest):
             # otherwise we invoke the task
             task_request = task_input  # type: TaskRequest
+            context.apply_task_meta(task_request)
             return self._invoke_task(context, task_request)
 
         else:
@@ -275,8 +275,7 @@ class FlinkTasks(object):
             raise ValueError(f'Unexpected type for task_input: {type(task_input)}')
 
     def _get_pipeline(self, context):
-        state = context.get_state()
-        pipeline_protos = state.get('pipeline', None)
+        pipeline_protos = context.pipeline_state.pipeline
 
         if pipeline_protos is not None:
             return _Pipeline.from_proto(pipeline_protos, self._serialiser)
@@ -306,11 +305,9 @@ class FlinkTasks(object):
         flink_task = self.get_task(task_request.type)
         fn = flink_task.run_async if flink_task.is_async else flink_task.run
 
-        return task_request, fn(task_request)
+        return task_request, fn(context, task_request)
 
     def _emit_result(self, context, task_input, task_result):
-        state = context.get_state()
-
         # either send a message to egress if reply_topic was specified
         if task_input.HasField('reply_topic'):
             context.pack_and_send_egress(topic=task_input.reply_topic, value=task_result)
@@ -323,18 +320,18 @@ class FlinkTasks(object):
         elif isinstance(task_input, TaskRequest):
             # if this was a task request it could be part of a pipeline - i.e. called from another tasks so
             # either call back to original caller (e.g. if this is a result of a retry)
-            if isinstance(task_input, TaskRequest) and 'original_caller_address' in state and 'original_caller_id' in state:
-                context.pack_and_send(state['original_caller_address'], state['original_caller_id'], task_result)
+            if context.task_state.original_caller_id != '':
+                context.pack_and_send(context.task_state.original_caller_address, context.task_state.original_caller_id, task_result)
 
             # or call back to our caller (if there is one)
             elif context.get_caller_id() is not None:
                 context.pack_and_reply(task_result)
 
     def _finalise_task_result(self, context, task_request, result_tuple):
-        task_result, task_exception, pipeline = result_tuple  # unpack
+        task_result, task_exception, pipeline, task_state = result_tuple  # unpack
 
-        if pipeline is not None:
-            pipeline.begin(context)
+        if pipeline is not None and not pipeline.is_empty():
+            pipeline.begin(context, task_request, task_state)
         else:
             if task_exception is not None:
                 if self._attempt_retry(context, task_request, task_exception):
@@ -350,27 +347,24 @@ class FlinkTasks(object):
             self._emit_result(context, task_request, return_value)
 
     def _attempt_retry(self, context, task_request, task_exception):
-        state = context.get_state()
-
         if task_exception.maybe_retry and task_exception.retry_policy is not None:
-
-            retry_count = state.get('retry_count', 0)
-            if retry_count >= task_exception.retry_policy.max_retries:
+            if context.task_state.retry_count >= task_exception.retry_policy.max_retries:
                 return False
 
             # save the original caller address and id for this task_request prior to calling back to ourselves which would overwrite
-            if not 'original_caller_id' in state or not 'original_caller_id' in state:
-                context.update_state({'original_caller_address': context.get_caller_address(), 'original_caller_id': context.get_caller_id()})
+            if context.task_state.original_caller_id == '':
+                context.task_state.original_caller_id = context.get_caller_id()
+                context.task_state.original_caller_address = context.get_caller_address()
 
             # remove previous task_request from state, increment retry count
             context.delete('task_request')
-            state['retry_count'] = retry_count + 1
+            context.task_state.retry_count += 1
 
             # send retry
             delay = timedelta(milliseconds=task_exception.retry_policy.delay_ms)
 
             if task_exception.retry_policy.exponential_back_off:
-                delay = timedelta(milliseconds=task_exception.retry_policy.delay_ms ^ retry_count)
+                delay = timedelta(milliseconds=task_exception.retry_policy.delay_ms * (2 ** context.task_state.retry_count))
 
             if delay:
                 context.pack_and_send_after(delay, context.get_address(), context.get_task_id(), task_request)
@@ -384,20 +378,25 @@ class FlinkTasks(object):
     def _fail(self, context, task_input, ex):
         task_exception = _create_task_exception(task_input, ex)
 
+        context.delete('task_result')
+        context.pack_and_save('task_exception', task_exception)
+
         # emit the error - either by replying to caller or sending some egress
         self._emit_result(context, task_input, task_exception)
 
 
 class _FlinkTask(object):
-    def __init__(self, fun, serialiser, retry_policy=None, with_state=False, is_fruitful=True, **kwargs):
+    def __init__(self, fun, serialiser, retry_policy=None, with_state=False, is_fruitful=True, with_context=False, **kwargs):
         self._fun = fun
         self._serialiser = serialiser
         self._retry_policy = retry_policy
         self._with_state = with_state
         self._is_fruitful = is_fruitful
+        self._with_context = with_context
 
         full_arg_spec = inspect.getfullargspec(fun)
         self._args = full_arg_spec.args
+        self._default_args = full_arg_spec.defaults
         self._num_args = len(self._args)
         self._accepts_varargs = full_arg_spec.varargs is not None
         self.is_async = inspect.iscoroutinefunction(fun)
@@ -406,79 +405,87 @@ class _FlinkTask(object):
         proto_types = _annotated_protos_for(fun)
         self._serialiser.register_proto_types(proto_types)
 
-    def run(self, task_request: TaskRequest):
+    def run(self, task_context: TaskContext, task_request: TaskRequest):
         task_result, task_exception, pipeline = None, None, None
 
+        task_args, kwargs, fn_state = self._to_task_args_and_kwargs(task_context, task_request)
+
         try:
             # run the flink task
-            task_args, kwargs, original_state = self._to_task_args_and_kwargs(task_request)
-
             fn_result = self._fun(*task_args, **kwargs)
 
-            pipeline, task_result = self._to_pipeline_or_task_result(task_request, fn_result, original_state)
+            pipeline, task_result, fn_state = self._to_pipeline_or_task_result(task_request, fn_result, fn_state)
 
         # we errored so return a task_exception instead
         except Exception as e:
             task_exception = self._to_task_exception(task_request, e)
 
-        return task_result, task_exception, pipeline
+        return task_result, task_exception, pipeline, fn_state
 
-    async def run_async(self, task_request: TaskRequest):
+    async def run_async(self, task_context: TaskContext, task_request: TaskRequest):
         task_result, task_exception, pipeline = None, None, None
 
+        task_args, kwargs, fn_state = self._to_task_args_and_kwargs(task_context, task_request)
+
         try:
             # run the flink task
-            task_args, kwargs, original_state = self._to_task_args_and_kwargs(task_request)
-
             fn_result = self._fun(*task_args, **kwargs)
 
             # await coro
             if asyncio.iscoroutine(fn_result):
                 fn_result = await fn_result
 
-            pipeline, task_result = self._to_pipeline_or_task_result(task_request, fn_result, original_state)
+            pipeline, task_result, fn_state = self._to_pipeline_or_task_result(task_request, fn_result, fn_state)
 
         # we errored so return a task_exception instead
         except Exception as e:
             task_exception = self._to_task_exception(task_request, e)
 
-        return task_result, task_exception, pipeline
+        return task_result, task_exception, pipeline, fn_state
 
-    def _to_pipeline_or_task_result(self, task_request, fn_result, original_state):
-        pipeline, task_result, fn_state = None, None, original_state
+    def _to_pipeline_or_task_result(self, task_request, fn_result, fn_state):
+        pipeline, task_result = None, None
 
-        if not self._is_fruitful:
+        # if single result then wrap in tuple as this is the maximal case
+        if not _is_tuple(fn_result):
+            fn_result = (fn_result,)
+
+        # if this task accesses state then we expect the first element in the result tuple
+        # to be the mutated state and the task results to be remainder
+        if self._with_state:
+
+            if len(fn_result) < 1:
+                raise ValueError('Expecting a tuple with at least the state as the first element')
+
+            fn_state = fn_result[0]
+            fn_result = fn_result[1:] if len(fn_result) > 1 else ()
+
+        # if a single element tuple remains then unpack back to single value
+        # so (8,) becomes 8 but (8,9) remains a tuple
+        fn_result = fn_result[0] if len(fn_result) == 1 else fn_result
+
+        # result of the task might be a Flink pipeline
+        if isinstance(fn_result, PipelineBuilder):
+            # this new pipeline which once complete will yield the result of the whole pipeline
+            # back to the caller as if it were a simple task
+            pipeline = fn_result.to_pipeline(self._serialiser, is_fruitful=self._is_fruitful)
             fn_result = ()
 
-        else:
-            # if single result then wrap in tuple as this is the maximal case
-            if not _is_tuple(fn_result):
-                fn_result = (fn_result,)
-
-            # if this task accesses state then we expect the first element in the result tuple
-            # to be the mutated state and the task results to be remainder
-            if self._with_state:
-                fn_state = fn_result[0]
-                fn_result = fn_result[1:] if len(fn_result) > 1 else ()
-
-            # if a single element tuple then unpack back to single value
-            # so (8,) becomes 8 but (8,9) remains a tuple
-            fn_result = fn_result[0] if len(fn_result) == 1 else fn_result
-
-            # result of the task might be a Flink pipeline
-            if isinstance(fn_result, PipelineBuilder):
-                pipeline = fn_result.to_pipeline()
-                fn_result = None
+        # drop the result if the task is marked as not fruitful
+        if not self._is_fruitful:
+            fn_result = ()
 
         task_result = _create_task_result(task_request)
         self._serialiser.serialise_result(task_result, fn_result, fn_state)
 
-        return pipeline, task_result
+        return pipeline, task_result, fn_state
 
     def _to_task_exception(self, task_request, ex):
         # use retry policy on task request first then fallback to task definition
-        task_parameters = self._serialiser.from_proto(task_request.parameters, {})
-        task_retry_policy = task_parameters.get('retry_policy', self._retry_policy)
+        task_retry_policy = task_request.retry_policy
+        if not task_request.HasField('retry_policy'):
+            task_retry_policy = self._retry_policy
+
         maybe_retry = False
 
         if task_retry_policy is not None:
@@ -493,7 +500,7 @@ class _FlinkTask(object):
 
         return task_exception
 
-    def _to_task_args_and_kwargs(self, task_request):
+    def _to_task_args_and_kwargs(self, task_context, task_request):
         args, kwargs, state = self._serialiser.deserialise_request(task_request)
 
         # listify
@@ -502,18 +509,34 @@ class _FlinkTask(object):
         else:
             args = [args]
 
-        # merge in args passed as kwargs e.g. fun1.continue_with(fun2, arg1=a, arg2=b)
-        args_in_kwargs = [(idx, arg, kwargs[arg]) for idx, arg in enumerate(self._args) if arg in kwargs]
-        for idx, arg, val in args_in_kwargs:
-            args.insert(idx, val)
-            del kwargs[arg]
-
         # add state as first argument if required by this task
         if self._with_state:
-            args = [state] + args
+            args.insert(0, state)
+        # add context if required by this task
+        if self._with_context:
+            args.insert(0, task_context)
 
-        return args, kwargs, state
+        resolved_args = []
+        # merge in args passed as kwargs e.g. fun1.continue_with(fun2, arg1=a, arg2=b)
+        for idx, arg in enumerate(self._args):
+            if arg in kwargs:
+                resolved_args.append(kwargs[arg])
+                del kwargs[arg]
+            elif self._default_args is not None and len(args) == 0:
+                default_arg_index = len(self._default_args) - len(self._args) + idx
+                if default_arg_index < 0:
+                    raise ValueError('Not enough args supplied')
+                resolved_args.append(self._default_args[default_arg_index])
+            else:
+                resolved_args.append(args.pop(0))
+        if self._accepts_varargs:
+            resolved_args.extend(args)
+        elif len(args) > 0:
+            raise ValueError('Too many args supplied')
+
+        return resolved_args, kwargs, state
 
 
-def in_parallel(entries: list):
-    return PipelineBuilder().append_group(entries)
+def in_parallel(entries: list, max_parallelism=None):
+    return PipelineBuilder().append_group(entries, max_parallelism)
+
