@@ -1,5 +1,6 @@
-from .messages_pb2 import TaskRequest, TaskResult, TaskException, ArgsAndKwargs
-from ._protobuf import _convert_from_proto, _convert_to_proto, _pack_any, _unpack_any, _parse_any_from_bytes
+from .messages_pb2 import TaskRequest, TaskResult, TaskException, ArgsAndKwargs, MapOfStringToAny, \
+    TupleOfAny
+from ._protobuf import _convert_from_proto, _convert_to_proto, _pack_any, _unpack_any
 from ._utils import _is_tuple
 from google.protobuf.any_pb2 import Any
 from google.protobuf.message import Message
@@ -66,7 +67,7 @@ class DefaultSerialiser(object):
 
     def deserialise_args_and_kwargs(self, request: Any):
         """
-        Deserialises Any into args and kwargs
+        Deserialises Any into Python args and kwargs
         
         :param request: the protobuf message
         :return: tuple of args and kwargs
@@ -82,18 +83,63 @@ class DefaultSerialiser(object):
 
         return args, kwargs
 
-    def serialise_request(self, task_request: TaskRequest, args, kwargs, state=None):
+    def to_args_and_kwargs(self, request: Any) -> ArgsAndKwargs:
+        """
+        Unpacks Any into Protobuf ArgsAndKwargs
+        
+        :param request: the protobuf message
+        :return: ArgsAndKwargs
+        """
+        unpacked = _convert_from_proto(request, self._known_proto_types)
+        if isinstance(unpacked, ArgsAndKwargs):
+            return unpacked
+        else:
+            args = TupleOfAny()
+            args.items.append(request)
+            return ArgsAndKwargs(args=args, kwargs=MapOfStringToAny())
+
+    @staticmethod
+    def merge_args_and_kwargs(task_result: Message, task_args_and_kwargs: ArgsAndKwargs) -> Any:
+        """
+        Merges args & kwargs passed explicity to a task entry in a pipeline with results from the previous task.
+        If there are no args & kwargs to merge then the result of the previous task is returned unchanged
+
+        :param task_result: the request payload proto
+        :param task_args_and_kwargs: the args and kwargs from the task entry to merge
+        :return: Any
+        """
+        args_to_merge = task_args_and_kwargs.args
+        kwargs = task_args_and_kwargs.kwargs
+
+        if not any(args_to_merge.items) and not any(kwargs.items):
+            return _pack_any(task_result)
+        
+        # task result may be a single proto in which case we have to wrap into TupleOfAny to be able to extend
+        if not isinstance(task_result, TupleOfAny):
+            args = TupleOfAny()
+            args.items.append(_pack_any(task_result))
+        else:
+            args = task_result
+
+        args.items.extend(args_to_merge.items)
+
+        return _pack_any(ArgsAndKwargs(args=args, kwargs=kwargs))
+
+    @staticmethod
+    def serialise_request(task_request: TaskRequest, request: Any, state=None, retry_policy=None):
         """
         Serialises args, kwargs and optional state into a TaskRequest
         
         :param task_request: the TaskRequest
-        :param args: task args
-        :param kwargs: task kwargs
+        :param request: request (proto format)
         :param optional state: task state
+        :param optional retry_policy: task retry policy
         """
-        request = self.serialise_args_and_kwargs(args, kwargs)
         task_request.request.CopyFrom(request)
         task_request.state.CopyFrom(_pack_any(_convert_to_proto(state)))
+        
+        if retry_policy:
+            task_request.retry_policy.CopyFrom(retry_policy)
 
     def deserialise_request(self, task_request: TaskRequest):
         """
@@ -130,17 +176,20 @@ class DefaultSerialiser(object):
 
         return result, state
 
-    def deserialise_response(self, task_result_or_exception: Union[TaskResult, TaskException]):
+    def unpack_response(self, task_result_or_exception: Union[TaskResult, TaskException]):
         """
-        Deserialises a TaskResult or TaskException back into result and state.
-        If a TaskException is provided as input then the return value will be (), TaskException.state
+        Unpacks a TaskResult or TaskException back into result and state.
+        If a TaskException is provided as input then the return value will be TupleOfAny, TaskException.state
         
         :param task_result: the TaskResult
         :return: tuple of result and state. 
         """
         if isinstance(task_result_or_exception, TaskResult):
-            return self.deserialise_result(task_result_or_exception)
+            task_result = _unpack_any(task_result_or_exception.result, self._known_proto_types)
+            task_state = task_result_or_exception.state
         elif isinstance(task_result_or_exception, TaskException):
-            return (), self.from_proto(task_result_or_exception.state)
+            task_result = TupleOfAny()
+            task_state = task_result_or_exception.state
         else:
             raise ValueError(f'task_result_or_exception was neither TaskResult or TaskException')
+        return task_result, task_state
