@@ -1,6 +1,7 @@
 import inspect
 from uuid import uuid4
 from typing import get_type_hints
+from itertools import chain
 from google.protobuf.message import Message
 from google.protobuf.any_pb2 import Any
 
@@ -28,6 +29,14 @@ def _try_next(iterator):
         return None
 
 
+def _try_peek(iterator):
+    try:
+        n = next(iterator)
+        return n, chain([n], iterator)
+    except StopIteration:
+        return None, iterator
+
+
 def _is_named_tuple(value):
     # duck test to see if a value is a NamedTuple and not just a tuple
     if not isinstance(value, tuple):
